"""Contains the interfaces that goals will implement."""
import abc
import typing as t


class Linter(abc.ABC):
    """The interface that all linters will implement."""

    name: str
    parallel_run: bool
    weight: int = 0

    def __init__(self, args: t.Optional[t.List[str]] = None):
        """Will initialize the linter.

        Args are passed through to the linter.
        """
        if args is None:
            args = []
        self._args = args

    @abc.abstractmethod
    def run(self):
        """Will run the linter.

        Depending on the linter this may have side effects
        """
        raise NotImplementedError

    @abc.abstractmethod
    def check(self):
        """Will run the linter in check mode.This should NEVER change any files."""
        raise NotImplementedError



"""
Backends for py_mono_tools.

A Backend is what takes the goals instructions and runs them. This could be the local system, docker, or others.
"""
from py_mono_tools.backends.docker import Docker  # noqa: F401
from py_mono_tools.backends.system import System  # noqa: F401



"""The system backend takes the goals instructions and runs them on the local system."""
import subprocess  # nosec B404
import typing as t

from py_mono_tools.backends.interface import Backend
from py_mono_tools.config import consts


class System(Backend):
    """Class to interact with the local system."""

    name: str = "system"

    def build(self, force_rebuild: bool = False):
        """Will do nothing for the system backend."""

    def purge(self):
        """Will do nothing for the system backend."""

    def run(self, args: t.List[str]):
        """Will run a command on the local system."""
        with subprocess.Popen(  # nosec B603
            args,
            cwd=consts.EXECUTED_FROM,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        ) as process:

            stdout_data, stderr_data = process.communicate()
        return process.returncode, stderr_data.decode("utf-8") + stdout_data.decode("utf-8")

    def interactive(self):
        """Will do nothing for the system backend."""

    def shutdown(self):
        """Will do nothing for the system backend."""



"""Store all the config/const data other files need."""
import logging
import os
import pathlib
import typing as t

from py_mono_tools.goals.interface import Linter


if t.TYPE_CHECKING:
    from py_mono_tools.backends.interface import Backend


# pylint: disable=too-few-public-methods
class Consts:
    """Used to store some "consts" that will be set at CLI runtime, then used in other modules."""

    EXECUTED_FROM: pathlib.Path = pathlib.Path(os.getcwd())  # pylint: disable=invalid-name
    CURRENT_BACKEND: t.Optional["Backend"] = None  # pylint: disable=invalid-name
    BACKENDS: t.Optional[t.Dict[str, t.Type["Backend"]]] = None  # pylint: disable=invalid-name
    CONF = None  # pylint: disable=invalid-name
    ALL_LINTERS: t.List[Linter] = []
    ALL_LINTER_NAMES: t.List[str] = []


consts = Consts()

BLACK = "\x1b[30m"
RED = "\x1b[31m"
GREEN = "\x1b[32m"
YELLOW = "\x1b[33m"
BLUE = "\x1b[34m"
MAGENTA = "\x1b[35m"
CYAN = "\x1b[36m"
WHITE = "\x1b[37m"
GRAY = "\x1b[38m"
RESET = "\x1b[0m"

FORMAT = "%(name)s-(%(filename)s:%(lineno)d)(%(process)d:%(thread)d) - %(levelname)s - %(asctime)s - %(message)s "


class ColorFormatting(logging.Formatter):
    """Used to colorize the log output."""

    FORMATS = {
        logging.DEBUG: GRAY + FORMAT + RESET,
        logging.INFO: GRAY + FORMAT + RESET,
        logging.WARNING: YELLOW + FORMAT + RESET,
        logging.ERROR: RED + FORMAT + RESET,
        logging.CRITICAL: RED + FORMAT + RESET,
    }

    def format(self, record):
        """Add formatting to the log output."""
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


logger = logging.getLogger("PMB")
logger.setLevel(logging.INFO)

stream = logging.StreamHandler()
stream.setLevel(logging.INFO)
stream.setFormatter(ColorFormatting())
logger.addHandler(stream)



"""Contains all the commands that the CLI can execute."""
import importlib
import importlib.machinery
import importlib.util
import inspect
import logging
import os
import pathlib
import sys
import typing as t

import click

from py_mono_tools.backends import Docker, System
from py_mono_tools.config import consts, logger
from py_mono_tools.goals import linters as linters_mod
from py_mono_tools.goals.interface import Linter


logger.info("Starting main")


def _find_linters():
    logger.debug("Finding linters")
    linter_classes = inspect.getmembers(linters_mod, inspect.isclass)
    linter_instances = [
        linter_class[1]()
        for linter_class in linter_classes
        if issubclass(linter_class[1], Linter) and linter_class[1] != Linter
    ]
    consts.ALL_LINTERS = linter_instances

    for linter in linter_instances:
        consts.ALL_LINTER_NAMES.append(linter.name)

    logger.debug("Found linters: %s", consts.ALL_LINTER_NAMES)


_find_linters()


def _load_conf():
    conf_location = f"{consts.EXECUTED_FROM}/CONF"
    logger.debug("CONF location: %s", conf_location)

    loader = importlib.machinery.SourceFileLoader("CONF", conf_location)
    spec = importlib.util.spec_from_loader(loader.name, loader)
    mod = importlib.util.module_from_spec(spec)  # type: ignore
    loader.exec_module(mod)

    # pylint: disable=invalid-name
    consts.CONF = mod


def _init_logger(verbose: bool):
    if verbose:
        logger.setLevel(logging.DEBUG)
        for handler in logger.handlers:
            handler.setLevel(logging.DEBUG)


def _set_absolute_path(absolute_path: str):
    logger.info("Overwriting execution root path: %s", absolute_path)

    # pylint: disable=invalid-name
    consts.EXECUTED_FROM = pathlib.Path(absolute_path).resolve()

    os.chdir(consts.EXECUTED_FROM.resolve())  # pylint: disable=invalid-name


def _set_relative_path(relative_path: str):
    logger.info("Overwriting execution root path: %s", relative_path)

    # pylint: disable=invalid-name
    consts.EXECUTED_FROM = consts.EXECUTED_FROM.joinpath(pathlib.Path(relative_path).resolve())
    os.chdir(consts.EXECUTED_FROM)


def _init_backend(_build_system: str):
    logger.debug("Initializing build system: %s", _build_system)

    # pylint: disable=invalid-name
    consts.BACKENDS = {
        Docker.name: Docker,
        System.name: System,
    }

    consts.CURRENT_BACKEND = consts.BACKENDS[_build_system]()


@click.group()
@click.option(
    "--backend",
    default=None,
    type=str,
    help="""
Default backend is \"system\".
This can be set via this flag, the BACKEND var in CONF, or defaulted to system.
""",
)
@click.option("--absolute_path", "-ap", default=None, type=click.Path())
@click.option("--relative_path", "-rp", default=None, type=click.Path())
@click.option("--verbose", "-v", default=False, is_flag=True)
def cli(backend, absolute_path, relative_path, verbose):
    """Py mono tool is a CLI tool that simplifies using python in a monorepo."""
    logger.info("Starting cli")

    if "--help" in sys.argv or "-h" in sys.argv:
        return

    _init_logger(verbose=verbose)

    logger.debug("Executed from: %s", consts.EXECUTED_FROM)
    logger.debug("Current backend: %s", consts.CURRENT_BACKEND)
    logger.debug("Backends: %s", consts.BACKENDS)

    if absolute_path is not None:
        _set_absolute_path(absolute_path=absolute_path)
    elif relative_path is not None:
        _set_relative_path(relative_path=relative_path)

    _load_conf()
    if backend is None:
        try:
            backend = consts.CONF.BACKEND
        except AttributeError:
            backend = "system"

    logger.info("Using backed: %s", backend)

    _init_backend(backend)


@cli.command()
@click.option("--force-rebuild", is_flag=True, default=False)
@click.option("--modules", multiple=True, type=t.List[str], default=["all"])
def build(force_rebuild, modules):
    """
    Run the build process for the specified build system.

    Docker is the default.
    """
    logger.info(modules)
    consts.CURRENT_BACKEND.build(force_rebuild=force_rebuild)


@cli.command()
def deploy():
    """Run the specified build and deploy in the specific CONF file."""
    raise NotImplementedError


@cli.command()
def init():
    """Initialize the current directory with the necessary files."""
    raise NotImplementedError


@cli.command()
@click.option("--check", is_flag=True, default=False)
@click.option(
    "--specific",
    "-s",
    multiple=True,
    default=[],
    help=f"""
    Specify one or more linters to run. NOTE: The linter MUST be listed in the respected CONF file.

    All Linters:
    {consts.ALL_LINTER_NAMES}
    """,
    shell_complete=consts.ALL_LINTER_NAMES,
)
@click.option("--fail_fast", "-ff", is_flag=True, default=False, help="Stop on first failure.")
@click.option("--show_success", is_flag=True, default=False, help="Show successful outputs")
@click.option(
    "--parallel",
    is_flag=True,
    default=False,
    help="""
    NOT IMPLEMENTED
    Runs all linters marked with parallel_run=True at the same time
    NOTE: All linters labeled as parallel_run=False will be run BEFORE ones marked as True.
    """,
)
@click.option(
    "--ignore_linter_weight", is_flag=True, default=False, help="Ignores linter weight and runs in the order in CONF."
)
def lint(
    check: bool,
    specific: t.List[str],
    fail_fast: bool,
    show_success: bool,
    parallel: bool,
    ignore_linter_weight: bool,
):  # pylint: disable=too-many-arguments
    """Run one or more Linters specified in the CONF file."""
    if parallel is True:
        raise NotImplementedError

    logger.info("Starting lint")

    linters: t.List[Linter] = consts.CONF.LINT  # type: ignore
    logger.debug("Linters: %s", linters)

    cleaned_spec = []
    for linter_name in specific:
        cleaned_spec.append(linter_name.strip().lower())
    if ignore_linter_weight is False:
        linters.sort(key=lambda x: x.weight, reverse=True)
    for linter in linters:
        if cleaned_spec and linter.name.strip().lower() not in cleaned_spec:
            logger.debug("Linters requested: %s", cleaned_spec)
            logger.debug("Skipping linter: %s", linter.name)
            continue

        logger.debug("Linting: %s", linter)
        if check is True:
            logs, return_code = linter.check()
        else:
            logs, return_code = linter.run()
        logger.info("Lint result: %s %s", linter.name, return_code)

        if show_success is False and return_code == 0:
            logger.debug("Skipping successful output")
            logger.debug(logs)
        else:
            logger.info(logs)

        if fail_fast is True and return_code != 0:
            logger.error("Linter %s failed with code %s", linter.name, return_code)
            sys.exit(1)

    logger.info("Linting complete")


@cli.group()
def new():
    """Create a new module with the specified name."""
    raise NotImplementedError


@new.command()
def migration():
    """Generate an Alembic Database migration."""
    raise NotImplementedError


@new.command()
def package():
    """Not implemented."""
    raise NotImplementedError


@cli.command()
def run():
    """Not implemented."""
    raise NotImplementedError


@cli.command()
def interactive():
    """Drop into an interactive session in your specified backend."""
    raise NotImplementedError


@cli.command()
def setup():
    """Not implemented."""
    raise NotImplementedError


@cli.command()
def test():
    """Run all the tests specified in the CONF file."""
    raise NotImplementedError


@cli.command()
def validate():
    """Not implemented."""
    raise NotImplementedError



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['py_mono_tools', 'py_mono_tools.backends', 'py_mono_tools.goals']

package_data = \
{'': ['*'], 'py_mono_tools': ['templates/*', 'templates/dockerfiles/*']}

install_requires = \
['Jinja2>=3,<4',
 'bandit>=1,<2',
 'black>=22,<23',
 'click>=8,<9',
 'flake8>=5,<6',
 'isort>=5,<6',
 'mypy>=0.971,<0.972',
 'pydocstringformatter>=0.7,<0.8',
 'pydocstyle[toml]>=6,<7',
 'pylint>=2,<3']

entry_points = \
{'console_scripts': ['pmt = py_mono_tools.main:cli',
                     'py_mono_tools = py_mono_tools.main:cli']}

setup_kwargs = {
    'name': 'py-mono-tools',
    'version': '0.0.2',
    'description': 'A CLI designed to make it easier to work in a python mono repo',
    'long_description': '# Python Mono Tools\n\n## Instillation\n\n`pip install py-mono-tools`\n\n## Usage\n\n`pmb lint`\n\n# More information\n\nFor more information, please go the GitHub page. https://peterhoburg.github.io/py-mono-tools/\n',
    'author': 'Peter Hoburg',
    'author_email': 'peterHoburg@users.noreply.github.com',
    'maintainer': 'Peter Hoburg',
    'maintainer_email': 'peterHoburg@users.noreply.github.com',
    'url': 'https://github.com/peterHoburg/py-mono-tools',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""The Docker backend takes the goals instructions and runs them in a docker container."""
import os
import subprocess  # nosec B404
import sys
import typing as t

from py_mono_tools.backends.interface import Backend
from py_mono_tools.config import consts, logger


# pylint: disable=R0801
class Docker(Backend):
    """Class to interact with a docker container."""

    name: str = "docker"

    def build(self, force_rebuild: bool = False):
        """Will shut down any running containers and builds a new one."""
        uid = os.getuid()

        self.shutdown()
        self._build(uid)

    def purge(self):
        """Will do nothing for the docker backend."""
        raise NotImplementedError

    def run(self, args: t.List[str]):
        """Will run a command in a docker container."""
        commands = [
            "docker",
            "run",
            "-it",
            "pmb_docker_backend",
        ]
        for arg in args:
            if isinstance(arg, str):
                commands.append("/opt/")
            else:
                commands.append(arg)

        logger.info("running command: %s", commands)

        self.build()
        with subprocess.Popen(  # nosec B603
            commands,
            cwd=consts.EXECUTED_FROM,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        ) as process:

            stdout_data, stderr_data = process.communicate()
        return process.returncode, stderr_data.decode("utf-8") + stdout_data.decode("utf-8")

    def interactive(self):
        """Not implemented for the docker backend."""
        raise NotImplementedError

    def shutdown(self):
        """Will shut down any running containers."""
        self._kill_all_containers()

    def _build(self, uid: int):
        env = {
            "DOCKER_BUILDKIT": "1",
            "BUILDKIT_PROGRESS": "plain",
        }
        with subprocess.Popen(  # nosec B607 B603
            [
                "docker",
                "build",
                "--build-arg",
                f"USER_UID={uid}",
                "-t",
                "pmb_docker_backend",
                ".",
            ],
            env=env,
            cwd=consts.EXECUTED_FROM,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        ) as process:
            stdout_data, stderr_data = process.communicate()

        logger.debug("Docker build stdout: \n%s", stdout_data.decode("utf-8"))
        logger.debug("Docker build stderr: \n%s", stderr_data.decode("utf-8"))
        if process.returncode != 0:
            logger.error(
                "Docker build failed: %s, stdout: %s stderr: %s",
                process.returncode,
                stdout_data.decode("utf-8"),
                stderr_data.decode("utf-8"),
            )
            sys.exit(1)

    def _kill_all_containers(self):
        try:
            subprocess.check_output(  # nosec B607 B603 B602
                "docker ps -q | xargs -r docker kill",
                shell=True,
                cwd=consts.EXECUTED_FROM,
            )
            # subprocess.check_output(  # nosec B607 B603
            #     ["docker-compose", "down", "--remove-orphans"],
            #     cwd=self._root_path,
            # )
        except subprocess.CalledProcessError:
            sys.exit(1)
